/*
Copyright © 2023-2025 Philipp Wolfer <phw@uploadedlobster.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

package deezer_test

import (
	"context"
	"net/http"
	"testing"

	"github.com/jarcoal/httpmock"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"go.uploadedlobster.com/scotty/internal/backends/deezer"
	"golang.org/x/oauth2"
)

func TestNewClient(t *testing.T) {
	token := oauth2.StaticTokenSource(&oauth2.Token{})
	client := deezer.NewClient(token)
	assert.IsType(t, deezer.Client{}, client)
}

func TestGetUserHistory(t *testing.T) {
	defer httpmock.DeactivateAndReset()

	token := oauth2.StaticTokenSource(&oauth2.Token{})
	client := deezer.NewClient(token)
	setupHTTPMock(t, client.HTTPClient.GetClient(),
		"https://api.deezer.com/user/me/history",
		"testdata/user-history.json")

	ctx := context.Background()
	result, err := client.UserHistory(ctx, 0, 2)
	require.NoError(t, err)

	assert := assert.New(t)
	assert.Equal(12, result.Total)
	require.Len(t, result.Tracks, 2)
	track1 := result.Tracks[0]
	assert.Equal(int64(1700753817), track1.Timestamp)
	assert.Equal("New Divide", track1.Track.Title)
	assert.Equal("Linkin Park", track1.Track.Artist.Name)
}

func TestGetUserTracks(t *testing.T) {
	defer httpmock.DeactivateAndReset()

	token := oauth2.StaticTokenSource(&oauth2.Token{})
	client := deezer.NewClient(token)
	setupHTTPMock(t, client.HTTPClient.GetClient(),
		"https://api.deezer.com/user/me/tracks",
		"testdata/user-tracks.json")

	ctx := context.Background()
	result, err := client.UserTracks(ctx, 0, 2)
	require.NoError(t, err)

	assert := assert.New(t)
	assert.Equal(4, result.Total)
	require.Len(t, result.Tracks, 2)
	track1 := result.Tracks[0]
	assert.Equal(int64(1700743848), track1.AddedAt)
	assert.Equal("Never Take Me Alive", track1.Track.Title)
	assert.Equal("Outland", track1.Track.Album.Title)
}

func setupHTTPMock(t *testing.T, client *http.Client, url string, testDataPath string) {
	httpmock.ActivateNonDefault(client)

	responder, err := httpmock.NewJsonResponder(200, httpmock.File(testDataPath))
	if err != nil {
		t.Fatal(err)
	}
	httpmock.RegisterResponder("GET", url, responder)
}
