/*
Copyright © 2023-2025 Philipp Wolfer <phw@uploadedlobster.com>

This file is part of Scotty.

Scotty is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.

Scotty is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Scotty. If not, see <https://www.gnu.org/licenses/>.
*/

package spotifyhistory

import (
	"context"
	"os"
	"path"
	"path/filepath"
	"slices"
	"sort"
	"time"

	"go.uploadedlobster.com/scotty/internal/config"
	"go.uploadedlobster.com/scotty/internal/i18n"
	"go.uploadedlobster.com/scotty/internal/models"
)

const historyFileGlob = "Streaming_History_Audio_*.json"

type SpotifyHistoryBackend struct {
	dirPath           string
	ignoreIncognito   bool
	ignoreSkipped     bool
	skippedMinSeconds int
}

func (b *SpotifyHistoryBackend) Name() string { return "spotify-history" }

func (b *SpotifyHistoryBackend) Options() []models.BackendOption {
	return []models.BackendOption{{
		Name:  "dir-path",
		Label: i18n.Tr("Directory path"),
		Type:  models.String,
	}, {
		Name:    "ignore-incognito",
		Label:   i18n.Tr("Ignore listens in incognito mode"),
		Type:    models.Bool,
		Default: "true",
	}, {
		Name:    "ignore-skipped",
		Label:   i18n.Tr("Ignore skipped listens"),
		Type:    models.Bool,
		Default: "false",
	}, {
		Name:    "ignore-min-duration-seconds",
		Label:   i18n.Tr("Minimum playback duration for skipped tracks (seconds)"),
		Type:    models.Int,
		Default: "30",
	}}
}

func (b *SpotifyHistoryBackend) InitConfig(config *config.ServiceConfig) error {
	b.dirPath = config.GetString("dir-path")
	b.ignoreIncognito = config.GetBool("ignore-incognito", true)
	b.ignoreSkipped = config.GetBool("ignore-skipped", false)
	b.skippedMinSeconds = config.GetInt("ignore-min-duration-seconds", 30)
	return nil
}

func (b *SpotifyHistoryBackend) ExportListens(ctx context.Context, oldestTimestamp time.Time, results chan models.ListensResult, progress chan models.TransferProgress) {
	files, err := filepath.Glob(path.Join(b.dirPath, historyFileGlob))
	p := models.TransferProgress{
		Export: &models.Progress{},
	}

	if err != nil {
		p.Export.Abort()
		progress <- p
		results <- models.ListensResult{Error: err}
		return
	}

	slices.Sort(files)
	fileCount := int64(len(files))
	p.Export.Total = fileCount
	for i, filePath := range files {
		if err := ctx.Err(); err != nil {
			results <- models.ListensResult{Error: err}
			p.Export.Abort()
			progress <- p
			return
		}

		history, err := readHistoryFile(filePath)
		if err != nil {
			results <- models.ListensResult{Error: err}
			p.Export.Abort()
			progress <- p
			return
		}
		listens := history.AsListenList(ListenListOptions{
			IgnoreIncognito:   b.ignoreIncognito,
			IgnoreSkipped:     b.ignoreSkipped,
			skippedMinSeconds: b.skippedMinSeconds,
		})
		sort.Sort(listens)
		results <- models.ListensResult{Items: listens}
		p.Export.Elapsed = int64(i)
		p.Export.TotalItems += len(listens)
		progress <- p
	}

	p.Export.Complete()
	progress <- p
}

func readHistoryFile(filePath string) (StreamingHistory, error) {
	file, err := os.Open(filePath)
	if err != nil {
		return nil, err
	}

	defer file.Close()
	history := StreamingHistory{}
	err = history.Read(file)
	if err != nil {
		return nil, err
	}

	return history, nil
}
