/*
Copyright © 2023-2025 Philipp Wolfer <phw@uploadedlobster.com>

This file is part of Scotty.

Scotty is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.

Scotty is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
Scotty. If not, see <https://www.gnu.org/licenses/>.
*/
package subsonic

import (
	"context"
	"net/http"
	"sort"
	"time"

	"github.com/supersonic-app/go-subsonic/subsonic"
	"go.uploadedlobster.com/mbtypes"
	"go.uploadedlobster.com/scotty/internal/config"
	"go.uploadedlobster.com/scotty/internal/i18n"
	"go.uploadedlobster.com/scotty/internal/models"
	"go.uploadedlobster.com/scotty/internal/version"
)

type SubsonicApiBackend struct {
	client   subsonic.Client
	password string
}

func (b *SubsonicApiBackend) Name() string { return "subsonic" }

func (b *SubsonicApiBackend) Options() []models.BackendOption {
	return []models.BackendOption{{
		Name:  "server-url",
		Label: i18n.Tr("Server URL"),
		Type:  models.String,
	}, {
		Name:  "username",
		Label: i18n.Tr("User name"),
		Type:  models.String,
	}, {
		Name:  "token",
		Label: i18n.Tr("Access token"),
		Type:  models.Secret,
	}}
}

func (b *SubsonicApiBackend) InitConfig(config *config.ServiceConfig) error {
	b.client = subsonic.Client{
		Client:     &http.Client{},
		BaseUrl:    config.GetString("server-url"),
		User:       config.GetString("username"),
		ClientName: version.AppName,
	}
	b.password = config.GetString("token")
	return nil
}

func (b *SubsonicApiBackend) ExportLoves(ctx context.Context, oldestTimestamp time.Time, results chan models.LovesResult, progress chan models.TransferProgress) {
	err := b.client.Authenticate(b.password)
	p := models.TransferProgress{
		Export: &models.Progress{},
	}
	if err != nil {
		p.Export.Abort()
		progress <- p
		results <- models.LovesResult{Error: err}
		return
	}

	starred, err := b.client.GetStarred2(map[string]string{})
	if err != nil {
		p.Export.Abort()
		progress <- p
		results <- models.LovesResult{Error: err}
		return
	}

	loves := b.filterSongs(starred.Song, oldestTimestamp)
	p.Export.Total = int64(len(loves))
	p.Export.Complete()
	progress <- p
	results <- models.LovesResult{Items: loves}
}

func (b *SubsonicApiBackend) filterSongs(songs []*subsonic.Child, oldestTimestamp time.Time) models.LovesList {
	loves := make(models.LovesList, 0, len(songs))
	for _, song := range songs {
		love := SongAsLove(*song, b.client.User)
		if love.Created.After(oldestTimestamp) {
			loves = append(loves, love)
		}
	}

	sort.Sort(loves)
	return loves
}

func SongAsLove(song subsonic.Child, username string) models.Love {
	recordingMBID := mbtypes.MBID(song.MusicBrainzID)
	love := models.Love{
		UserName:      username,
		Created:       song.Starred,
		RecordingMBID: recordingMBID,
		Track: models.Track{
			TrackName:     song.Title,
			ReleaseName:   song.Album,
			ArtistNames:   []string{song.Artist},
			TrackNumber:   song.Track,
			DiscNumber:    song.DiscNumber,
			RecordingMBID: recordingMBID,
			Tags:          []string{},
			AdditionalInfo: map[string]any{
				"subsonic_id": song.ID,
			},
			Duration: time.Duration(song.Duration * int(time.Second)),
		},
	}

	if len(song.Genres) > 0 {
		genres := make([]string, 0, len(song.Genres))
		for _, genre := range song.Genres {
			genres = append(genres, genre.Name)
		}
		love.Track.Tags = genres
	} else if song.Genre != "" {
		love.Track.Tags = []string{song.Genre}
	}

	return love
}
